from pathlib import Path
from typing import Union

from gspread import service_account, service_account_from_dict


class Connection:
    def __init__(
        self, credentials_path: Union[Path, str] = None, credentials_dict: dict = None
    ):
        if not credentials_path and credentials_dict:
            raise ValueError("credentials path or dict is required")

        self.client = (
            service_account(credentials_path)
            if credentials_path
            else service_account_from_dict(credentials_dict)
        )



from dataclasses import dataclass

from gspread_formatting import CellFormat, Color, TextFormat


@dataclass
class Style:
    bg_color: str = "#ffffff"
    text_color: str = "#000000"
    horizontal_align: str = "left"
    vertical_align: str = "middle"
    raw: CellFormat = None
    bold: bool = False
    italic: bool = False
    strikethrough: bool = False

    def __post_init__(self):
        if not self.raw:
            self.bg_color = Color.fromHex(self.bg_color)
            self.text_color = Color.fromHex(self.text_color)
            self.horizontal_align = self.horizontal_align.upper()
            self.vertical_align = self.vertical_align.upper()

            self.raw = CellFormat(
                backgroundColor=self.bg_color,
                textFormat=TextFormat(
                    foregroundColor=self.text_color,
                    bold=self.bold,
                    italic=self.italic,
                    strikethrough=self.strikethrough,
                ),
                horizontalAlignment=self.horizontal_align,
                verticalAlignment=self.vertical_align,
            )



def next_char(start: str, count: int) -> str:
    next_code = ord(start) + count
    if next_code <= ord("Z"):
        return chr(next_code)
    else:
        return chr(next_code - 26)


def to_range(start: str = "A", count: int = 0, index: int = 1):
    return f"{start}{index}:{next_char(start, count)}{index}"



from .cell import Cell
from .connection import Connection
from .row import Row
from .sheet import Sheet
from .style import Style
from .tab import Tab

__all__ = ["Sheet", "Row", "Cell", "Style", "Tab", "Connection"]



from typing import TYPE_CHECKING, List, Union

from gspread_formatting import format_cell_range, CellFormat

from .cell import Cell
from .style import Style
from .utils import next_char, to_range

if TYPE_CHECKING:
    from .tab import Tab


class Row(list):
    def __init__(self, items, tab: "Tab", index: int):
        self.tab = tab
        self.row_index = index
        self.items = self.convert_to_cell(items)
        self.range = to_range("A", len(self.items) - 1, index=self.row_index)
        super().__init__(self.items)

    def convert_to_cell(self, items):
        def to_cell(item):
            i, value = item
            return Cell(value, self.tab, next_char("A", i), self.row_index, self)

        converted: List[Cell] = list(map(to_cell, enumerate(items)))
        return converted

    def clear(self) -> None:
        self.tab.sheet.values_clear(self.range)
        self.items = []
        super().__init__([])

    def update(self, values: list, *args, **kwargs):
        self.range = to_range("A", len(values) - 1, index=self.row_index)
        self.tab.update(self.range, [values], *args, **kwargs)
        self.items = self.convert_to_cell(values)
        super().__init__(self.items)

    def style(self, obj: Union[Style, CellFormat]):
        style = obj.raw if isinstance(obj, Style) else obj
        format_cell_range(self.tab, self.range, style)

    def refetch(self):
        values = self.tab.values()
        row = values[self.row_index - 1]
        self.items = self.convert_to_cell(row)
        super().__init__(self.items)

    def append_cell(self, value, *args, **kwargs):
        new_values = value if isinstance(value, list) else [value]
        new_data = [*self.items, *new_values]
        self.update(new_data, *args, **kwargs)



from dataclasses import dataclass
from typing import List

from gspread import Worksheet
from gspread.utils import ValueRenderOption

from .cell import Cell
from .row import Row

render_formats = {
    "formatted": ValueRenderOption.formatted,
    "unformatted": ValueRenderOption.unformatted,
    "formula": ValueRenderOption.formula,
}


@dataclass
class Tab(Worksheet):
    def __init__(self, sheet, *args, **kwargs):
        self.sheet = sheet
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"<Tab title='{self.title}' id={self.id}>"

    def values(self, **kwargs) -> List[Row]:
        rows = self.get_values(**kwargs)

        for i, row in enumerate(rows):
            rows[i] = Row(row, tab=self, index=i + 1)

        return rows

    def get_row(self, serial_no: int):
        values = self.values()
        return values[serial_no - 1]

    def get_cell(self, cell_name: str, render_option: str = "formatted", **kwargs):
        data = self.get(
            cell_name,
            value_render_option=render_formats.get(render_option, "formatted"),
            **kwargs,
        )[0][0]

        return Cell(
            data, tab=self, label=cell_name[0], row_index=int(cell_name[1:]), row=None
        )

    def append(self, data: list):
        self.append_row(data)



from dataclasses import dataclass

from gspread import Client, Spreadsheet

from .connection import Connection
from .tab import Tab


@dataclass
class Sheet(Spreadsheet):
    sheet_name: str
    folder_id: str = None
    connection: Connection = None
    sheet: Spreadsheet = None
    _properties: dict = None

    def __post_init__(self):
        sheet = self.connection.client.open(self.sheet_name, folder_id=self.folder_id)
        self.client: Client = sheet.client
        self._properties: dict = sheet._properties  # noqa

    def get_tab(self, tab_name: str) -> Tab:
        _tab = self.worksheet(tab_name)
        return Tab(spreadsheet=self, properties=_tab._properties, sheet=self)  # noqa

    def tabs(self, exclude_hidden: bool = False):
        sheets = self.worksheets(exclude_hidden=exclude_hidden)
        return [
            Tab(spreadsheet=self, properties=tab._properties, sheet=self)  # noqa
            for tab in sheets
        ]
